#include <stdio.h>
#include <avr/io.h>
#include <avr/interrupt.h>
#include <util/delay.h>

extern "C" {
    #include "uartlibrary/uart.h"
    #include "onewire.h"
    #include "ds18x20.h"
}

int uart_putchar( char c, FILE *stream ) {
    uart_putc( c );
    return 0;
}

int uart_getchar( FILE *stream ) {
    return uart_getc();
}

FILE uart;

// arduino      atmega328 pin   port/bit
// D0           2               PD0             RX
// D1           3               PD1             TX
// D2           4               PD2             7seg    cathode
// D3           5               PD3             7seg    cathode
// D4           6               PD4             7seg    cathode
// D5           11              PD5             7seg    cathode
// D6           12              PD6             7seg    cathode
// D7           13              PD7             7seg    cathode
// D8           14              PB0             7seg    cathode
// D9           15              PB1             7seg    anode
// D10          16              PB2             7seg    anode
// D11          17              PB3             7seg    anode
// D12          18              PB4             oneWire
// D13          19              PB5             control
// A0           23              PC0
// A1           24              PC1
// A2           25              PC2
// A3           26              PC3
// A4           27              PC4
// A5           28              PC5

const unsigned int MAXSENSORS=8;
uint8_t gSensorIDs[MAXSENSORS][OW_ROMCODE_SIZE];

uint8_t search_sensors(void)
{
    uint8_t i;
    uint8_t id[OW_ROMCODE_SIZE];
    uint8_t diff, nSensors;

    nSensors = 0;

    for ( diff = OW_SEARCH_FIRST;
          diff != OW_LAST_DEVICE && nSensors < MAXSENSORS ; ) {
        DS18X20_find_sensor( &diff, &id[0] );

        if( diff == OW_PRESENCE_ERR ) {
            fprintf( &uart, "No sensor found\r\n" );
            break;
        }

        if( diff == OW_DATA_ERR ) {
            fprintf( &uart, "Bus Error while scanning for DS18X20\r\n" );
            break;
        }

        for (i=0;i<OW_ROMCODE_SIZE;i++)
            gSensorIDs[nSensors][i]=id[i];

        nSensors++;
    }

    return nSensors;
}

void seg7_setup() {
    // bits 2--7
    DDRD |= 0x3f << 2;
    DDRB |= 0xf;

    // Turn on timer0.
    TCCR0B = _BV( CS01 ) | _BV( CS00 );
    // enable the overflow interrupt
    TIMSK0 |= _BV( TOIE0 );
}

enum {
    SEG7_A,
    SEG7_B,
    SEG7_C,
    SEG7_D,
    SEG7_E,
    SEG7_F,
    SEG7_G
};
//                                   A  B  C  D  E  F  G
static const bool seg7_digit_0[] = { 1, 1, 1, 1, 1, 1, 0 };
static const bool seg7_digit_1[] = { 0, 1, 1, 0, 0, 0, 0 };
static const bool seg7_digit_2[] = { 1, 1, 0, 1, 1, 0, 1 };
static const bool seg7_digit_3[] = { 1, 1, 1, 1, 0, 0, 1 };
static const bool seg7_digit_4[] = { 0, 1, 1, 0, 0, 1, 1 };
static const bool seg7_digit_5[] = { 1, 0, 1, 1, 0, 1, 1 };
static const bool seg7_digit_6[] = { 1, 0, 1, 1, 1, 1, 1 };
static const bool seg7_digit_7[] = { 1, 1, 1, 0, 0, 0, 0 };
static const bool seg7_digit_8[] = { 1, 1, 1, 1, 1, 1, 1 };
static const bool seg7_digit_9[] = { 1, 1, 1, 1, 0, 1, 1 };
static const bool seg7_digit_blank[] = { 0, 0, 0, 0, 0, 0, 0 };
static const bool *seg7_digit_map[] = {
        seg7_digit_0, seg7_digit_1, seg7_digit_2, seg7_digit_3, seg7_digit_4,
        seg7_digit_5, seg7_digit_6, seg7_digit_7, seg7_digit_8, seg7_digit_9,
        seg7_digit_blank
        };
static volatile unsigned char *segment_port[] = {
        // A        B      C      D       E       F       G
        &PORTD, &PORTB, &PORTD, &PORTD, &PORTD, &PORTD, &PORTD
    };
static const uint8_t segment_bit[] = {
        //A   B     C     D     E     F     G
        0x80, 0x01, 0x10, 0x08, 0x04, 0x40, 0x20
    };
uint8_t seg7_digits[] = { 10, 10, 10 };
void seg7_update()
{
    static uint8_t state = 0;

    // All off.
    // PORTD |= 0x3f << 2;
    // PORTB |= _BV(0);
    // PORTB &= ~( _BV( 1 ) | _BV( 2 ) | _BV( 3 ) );

    // Turn off the old segment.
    *(segment_port[ state ]) |= segment_bit[ state ];

    state++;
    if ( state >= 7 ) {
        state = 0;
    }

    // Turn on the new segment.
    *(segment_port[ state ]) &= ~segment_bit[ state ];

    // For each digit, enable the digit if the current segment needs to be
    // shown. Otherwise disable it.
    if ( seg7_digit_map[ seg7_digits[ 0 ] ][ state ] )
        PORTB |= _BV( 1 );
    else
        PORTB &= ~_BV( 1 );
    if ( seg7_digit_map[ seg7_digits[ 1 ] ][ state ] )
        PORTB |= _BV( 2 );
    else
        PORTB &= ~_BV( 2 );
    if ( seg7_digit_map[ seg7_digits[ 2 ] ][ state ] )
        PORTB |= _BV( 3 );
    else
        PORTB &= ~_BV( 3 );
}

void seg7_set_number( unsigned int number )
{
    seg7_digits[ 0 ] = number % 10;
    if ( number >= 10 ) {
        seg7_digits[ 1 ] = ( number / 10 ) % 10;
    } else {
        seg7_digits[ 1 ] = 10;
    }
    if ( number >= 100 ) {
        seg7_digits[ 2 ] = ( number / 100 ) % 10;
    } else {
        seg7_digits[ 2 ] = 10;
    }
}

ISR(TIMER0_OVF_vect) {
    seg7_update();
}

// in Fahrenheit
static const int goal_temperature = 210;

int main( int argc, char **argv )
{
    fdev_setup_stream(&uart, uart_putchar, uart_getchar, _FDEV_SETUP_RW);
    uart_init( UART_BAUD_SELECT( 9600, F_CPU) ); 

    sei();

    DDRB |= _BV(5);
    seg7_setup();

    unsigned int nSensors = search_sensors();
    fprintf( &uart, "%d DS18X20 Sensor(s) available:\r\n", nSensors );

    while ( 1 ) {
        int temperature = 0;

        for ( unsigned int i = 0; i < nSensors; i++ ) {
            fprintf( &uart, "Sensor #%d: id=", i );
            for ( unsigned int j = 0; j < OW_ROMCODE_SIZE; j++ ) {
                fprintf( &uart, "%02x", gSensorIDs[i][j] );
            }

            if ( DS18X20_get_power_status( &gSensorIDs[i][0] ) ==
                    DS18X20_POWER_PARASITE ) {
                fprintf( &uart, ", parasite" );
            }

            if ( DS18X20_start_meas( DS18X20_POWER_PARASITE, 
                        &gSensorIDs[i][0] ) != DS18X20_OK ) {
                fprintf( &uart,
                         " failed to start measurement (short circuit?)\r\n");
                continue;
            }

            _delay_ms(DS18B20_TCONV_12BIT);
            int16_t measurement;
            if ( DS18X20_read_measurement( gSensorIDs[i], &measurement) !=
                    DS18X20_OK ) {
                fprintf( &uart, " failed to read measurement\r\n" );
                continue;
            }

            int16_t celsius;
            uint8_t centi_celsius;
            DS18X20_measurement_to_celsius(
                    measurement, &celsius, &centi_celsius );
            fprintf( &uart, ", %d.%02dC", celsius, centi_celsius );

            int16_t fahrenheit;
            uint8_t centi_fahrenheit;
            DS18X20_measurement_to_fahrenheit(
                    measurement, &fahrenheit, &centi_fahrenheit );
            fprintf( &uart, ", %d.%02dF", fahrenheit, centi_fahrenheit );
            temperature = fahrenheit;

            fprintf( &uart, "\r\n" );
        }

        seg7_set_number( temperature );
        if ( temperature < goal_temperature ) {
            PORTB |= _BV(5);
        } else if ( temperature > goal_temperature ) {
            PORTB &= ~_BV(5);
        }

        _delay_ms(1000);
    }

    return 0;
}
